from .FactorizedJetCorrector import FactorizedJetCorrector
from .JetResolution import JetResolution
from .JetResolutionScaleFactor import JetResolutionScaleFactor
from .JetCorrectionUncertainty import JetCorrectionUncertainty

from ..analysis_objects.JaggedCandidateArray import JaggedCandidateArray

import numpy as np
from uproot_methods import TLorentzVectorArray
from copy import deepcopy
from pdb import set_trace

_signature_map = {'JetPt': 'pt',
                  'JetEta': 'eta',
                  'Rho': 'rho',
                  'JetA': 'area'
                 }


def _update_jet_ptm(corr, jet, fromRaw=False):
    """
        This is a hack to update the jet pt and jet mass in place
        as we apply corrections and smearings.
    """
    if fromRaw:
        jet._content._contents['__fast_pt'] = corr * jet.ptRaw.content
        jet._content._contents['__fast_mass'] = corr * jet.massRaw.content
    else:
        jet._content._contents['__fast_pt'] = corr * jet.pt.content
        jet._content._contents['__fast_mass'] = corr * jet.mass.content


# the class below does use hacks of JaggedCandidateArray to achieve the desired behavior
# no monkey patches though
class JetTransformer(object):
    """
        This class is a columnar implementation of the the standard recipes for apply JECs, and
        the various scale factors and uncertainties therein.
           - Only the stochastic smearing method is implemented at the moment.
        It uses the FactorizedJetCorrector, JetResolution, JetResolutionScaleFactor, and
        JetCorrectionUncertainty classes to calculate the ingredients for the final updated jet
        object, which will be modified in place.
        The jet object must be a "JaggedCandidateArray" and have the additional properties:
           - ptRaw
           - massRaw
        These will be used to reset the jet pT and mass, and then calculate the updated pTs and
        masses for various corrections and smearings.
        You can use this class like:
        xformer = JetTransformer(name1=corrL1,...)
        xformer.transform(jet)
    """
    def __init__(self, jec=None, junc=None, jer=None, jersf=None):
        if jec is None:
            raise Exception('JetTransformer must have "jec" specified as an argument!')
        if not isinstance(jec, FactorizedJetCorrector):
            raise Exception('JetTransformer needs a FactorizedJetCorrecter passed as "jec"' +
                            ' got object of type {}'.format(type(jec)))
        self._jec = jec

        if junc is not None and not isinstance(junc, JetCorrectionUncertainty):
            raise Exception('"junc" must be of type "JetCorrectionUncertainty"' +
                            ' got {}'.format(type(junc)))
        self._junc = junc

        if (jer is None) != (jersf is None):
            raise Exception('Cannot apply JER-SF without an input JER, and vice-versa!')

        if jer is not None and not isinstance(jer, JetResolution):
            raise Exception('"jer" must be of type "JetResolution"' +
                            ' got {}'.format(type(jer)))
        self._jer = jer

        if jersf is not None and not isinstance(jersf, JetResolutionScaleFactor):
            raise Exception('"jersf" must be of type "JetResolutionScaleFactor"' +
                            ' got {}'.format(type(jersf)))
        self._jersf = jersf

    @property
    def uncertainties(self):
        return self._junc.levels if self._junc is not None else []

    def transform(self, jet, met=None):
        """
            precondition - jet is a JaggedCandidateArray with additional attributes:
                             - 'ptRaw'
                             - 'massRaw'
            xformer = JetTransformer(name1=corrL1,...)
            xformer.transform(jet)
            postcondition - jet.pt, jet.mass, jet.p4 are updated to represent the corrected jet
                            based on the input correction set
        """
        if not isinstance(jet, JaggedCandidateArray):
            raise Exception('Input data must be a JaggedCandidateArray!')
        if ('ptRaw' not in jet.columns or 'massRaw' not in jet.columns):
            raise Exception('Input JaggedCandidateArray must have "ptRaw" & "massRaw"!')

        if met is not None:
            if 'p4' not in met.columns:
                raise Exception('Input met must have a p4 column!')
            if not isinstance(met['p4'], TLorentzVectorArray):
                raise Exception('Met p4 must be a TLorentzVectorArray!')

        initial_p4 = jet['p4'].copy()  # keep a copy for fixing met
        # initialize the jet momenta to raw values
        _update_jet_ptm(1.0, jet, fromRaw=True)

        # below we work in numpy arrays, JaggedCandidateArray knows how to convert them
        args = {key: getattr(jet, _signature_map[key]).content for key in self._jec.signature}
        jec = self._jec.getCorrection(**args)

        _update_jet_ptm(jec, jet, fromRaw=True)

        juncs = None
        if self._junc is not None:
            args = {key: getattr(jet, _signature_map[key]).content for key in self._junc.signature}
            juncs = self._junc.getUncertainty(**args)

        # if there's a jer and sf to apply we have to update the momentum too
        # right now only use stochastic smearing
        if self._jer is not None and self._jersf is not None:
            args = {key: getattr(jet, _signature_map[key]).content for key in self._jer.signature}
            jer = self._jer.getResolution(**args)

            args = {key: getattr(jet, _signature_map[key]).content for key in self._jersf.signature}
            jersf = self._jersf.getScaleFactor(**args)

            jersmear = jer * np.random.normal(size=jer.size)
            jsmear_cen = 1. + np.sqrt(jersf[:, 0]**2 - 1.0) * jersmear
            jsmear_up = 1. + np.sqrt(jersf[:, 1]**2 - 1.0) * jersmear
            jsmear_down = 1. + np.sqrt(jersf[:, -1]**2 - 1.0) * jersmear

            # need to apply up and down jer-smear before applying central correction
            jet.add_attributes(pt_jer_up=jsmear_up * jet.pt.content,
                               mass_jer_up=jsmear_up * jet.mass.content,
                               pt_jer_down=jsmear_down * jet.pt.content,
                               mass_jer_down=jsmear_down * jet.mass.content)
            # finally, update the central value
            _update_jet_ptm(jsmear_cen, jet)

        # have to apply central jersf before calculating junc
        if self._junc is not None:
            for name, values in juncs:
                jet.add_attributes(**{
                    'pt_{0}_up'.format(name): values[:, 0] * jet.pt.content,
                    'mass_{0}_up'.format(name): values[:, 0] * jet.mass.content,
                    'pt_{0}_down'.format(name): values[:, 1] * jet.pt.content,
                    'mass_{0}_down'.format(name): values[:, 1] * jet.mass.content
                })

        # hack to update the jet p4, we have the fully updated pt and mass here
        jet._content._contents['p4'] = TLorentzVectorArray.from_ptetaphim(jet.pt.content,
                                                                          jet.eta.content,
                                                                          jet.phi.content,
                                                                          jet.mass.content)

        if met is None:
            return
        # set MET values
        new_x = met['p4'].x - (initial_p4.x - jet['p4'].x).sum()
        new_y = met['p4'].y - (initial_p4.y - jet['p4'].y).sum()
        met.base['p4'] = TLorentzVectorArray.from_ptetaphim(
            np.sqrt(new_x**2 + new_y**2), 0,
            np.arctan2(new_y, new_x), 0
        )
        if 'MetUnclustEnUpDeltaX' in met.columns:
            px_up = met['p4'].x + met['MetUnclustEnUpDeltaX']
            py_up = met['p4'].y + met['MetUnclustEnUpDeltaY']
            met.base['pt_UnclustEn_up'] = np.sqrt(px_up**2 + py_up**2)
            met.base['phi_UnclustEn_up'] = np.arctan2(py_up, px_up)

            px_down = met['p4'].x - met['MetUnclustEnUpDeltaX']
            py_down = met['p4'].y - met['MetUnclustEnUpDeltaY']
            met.base['pt_UnclustEn_down'] = np.sqrt(px_down**2 + py_down**2)
            met.base['phi_UnclustEn_down'] = np.arctan2(py_down, px_down)

        if self._junc is not None:
            jets_sin = np.sin(jet['p4'].phi)
            jets_cos = np.cos(jet['p4'].phi)
            for name, _ in juncs:
                for shift in ['up', 'down']:
                    px = met['p4'].x - (initial_p4.x - jet['pt_{0}_{1}'.format(name, shift)] * jets_cos).sum()
                    py = met['p4'].y - (initial_p4.y - jet['pt_{0}_{1}'.format(name, shift)] * jets_sin).sum()
                    met.base['pt_{0}_{1}'.format(name, shift)] = np.sqrt(px**2 + py**2)
                    met.base['phi_{0}_{1}'.format(name, shift)] = np.arctan2(py, px)



from fnal_column_analysis_tools.lookup_tools.lookup_base import lookup_base

from fnal_column_analysis_tools.util import awkward
from fnal_column_analysis_tools.util import numpy as np
from copy import deepcopy


def masked_bin_eval(dim1_indices, dimN_bins, dimN_vals):
    dimN_indices = np.empty_like(dim1_indices)
    for i in np.unique(dim1_indices):
        idx = np.where(dim1_indices == i)
        dimN_indices[idx] = np.clip(np.searchsorted(dimN_bins[i],
                                                    dimN_vals[idx],
                                                    side='right') - 1,
                                    0, len(dimN_bins[i]) - 2)
    return dimN_indices


class jersf_lookup(lookup_base):
    """
        This class defines a lookup table for jet energy resolution scale factors.
        The uncertainty values can be looked up with a call as follows:
        jersf_lut = jersf_lookup()
        SFs = jersf_lut(JetProperty1=jet.property1,...)
        "SFs" will be of the same shape as the input jet properties.
        The list of required jet properties are given in jersf_lut.signature
    """
    def __init__(self, formula, bins_and_orders, clamps_and_vars, parms_and_orders):
        """
            The constructor takes the output of the "convert_jersf_txt_file"
            text file converter, which returns a formula, bins, and values.
        """
        super(jersf_lookup, self).__init__()
        self._dim_order = bins_and_orders[1]
        self._bins = bins_and_orders[0]
        self._eval_vars = clamps_and_vars[2]
        self._eval_clamp_mins = clamps_and_vars[0]
        self._eval_clamp_maxs = clamps_and_vars[1]
        self._parm_order = parms_and_orders[1]
        self._parms = parms_and_orders[0]
        self._formula_str = formula
        self._formula = None
        if formula != 'None':
            raise Exception('jet energy resolution scale factors have no formula!')

        for binname in self._dim_order[1:]:
            binsaslists = self._bins[binname].tolist()
            self._bins[binname] = [np.array(bins) for bins in binsaslists]

        # get the jit to compile if we've got more than one bin dim
        if len(self._dim_order) > 1:
            masked_bin_eval(np.array([0]), self._bins[self._dim_order[1]], np.array([0.0]))

        self._signature = deepcopy(self._dim_order)
        for eval in self._eval_vars:
            if eval not in self._signature:
                self._signature.append(eval)
        self._dim_args = {self._dim_order[i]: i for i in range(len(self._dim_order))}
        self._eval_args = {}
        for i, argname in enumerate(self._eval_vars):
            self._eval_args[argname] = i + len(self._dim_order)
            if argname in self._dim_args.keys():
                self._eval_args[argname] = self._dim_args[argname]

    def _evaluate(self, *args):
        """ SFs = f(args) """
        bin_vals = {argname: args[self._dim_args[argname]] for argname in self._dim_order}
        eval_vals = {argname: args[self._eval_args[argname]] for argname in self._eval_vars}

        # lookup the bins that we care about
        dim1_name = self._dim_order[0]
        dim1_indices = np.clip(np.searchsorted(self._bins[dim1_name],
                                               bin_vals[dim1_name],
                                               side='right') - 1,
                               0, self._bins[dim1_name].size - 2)
        bin_indices = [dim1_indices]
        for binname in self._dim_order[1:]:
            bin_indices.append(masked_bin_eval(bin_indices[0], self._bins[binname],
                                               bin_vals[binname]))
        bin_tuple = tuple(bin_indices)

        # get clamp values and clip the inputs
        eval_values = []
        for eval_name in self._eval_vars:
            clamp_mins = self._eval_clamp_mins[eval_name][bin_tuple]
            clamp_maxs = self._eval_clamp_maxs[eval_name][bin_tuple]
            eval_values.append(np.clip(eval_vals[eval_name], clamp_mins, clamp_maxs))

        # get parameter values
        parm_values = self._parms[bin_tuple]

        return parm_values

    @property
    def signature(self):
        """ the list of all needed jet properties to be passed as kwargs to this lookup """
        return self._signature

    def __repr__(self):
        out = 'binned dims   : %s\n' % (self._dim_order)
        out += 'eval vars     : %s\n' % (self._eval_vars)
        out += 'return format : %s\n' % (self._parm_order)
        out += 'formula       : %s\n' % (self._formula_str)
        out += 'signature     : (%s)\n' % (','.join(self._signature))
        return out



from ..lookup_tools.jec_uncertainty_lookup import jec_uncertainty_lookup
import warnings
import re
from fnal_column_analysis_tools.util import awkward
from fnal_column_analysis_tools.util import numpy as np
from copy import deepcopy


def _checkConsistency(against, tocheck):
    if against is None:
        against = tocheck
    else:
        if against != tocheck:
            raise Exception('Corrector for {} is mixed'
                            'with correctors for {}!'.format(tocheck, against))
    return tocheck


class JetCorrectionUncertainty(object):
    """
        This class is a columnar implementation of the JetCorrectionUncertainty tool in
        CMSSW and FWLite. It calculates the jet energy scale uncertainty for a corrected jet
        in a given binning.
        You can use this class as follows:
        jcu = JetCorrectionUncertainty(name1=corrL1,...)
        jetUncs = jcu(JetParameter1=jet.parameter1,...)
    """
    def __init__(self, **kwargs):
        """
            You construct a JetCorrectionUncertainty by passing in a dict of names and functions.
            Names must be formatted as '<campaign>_<dataera>_<datatype>_<level>_<jettype>'.
        """
        jettype = None
        levels = []
        funcs = []
        datatype = None
        campaign = None
        dataera = None
        for name, func in kwargs.items():
            if not isinstance(func, jec_uncertainty_lookup):
                raise Exception('{} is a {} and not a jec_uncertainty_lookup!'.format(name,
                                                                                      type(func)))
            info = name.split('_')
            if len(info) == 6:  # this is when we are using split sources
                lvl = info.pop()
                info[3] = lvl

            if len(info) != 5:
                raise Exception('Corrector name is not properly formatted!')

            campaign = _checkConsistency(campaign, info[0])
            dataera = _checkConsistency(dataera, info[1])
            datatype = _checkConsistency(datatype, info[2])
            levels.append(info[3].replace('Uncertainty', 'jes'))  # use a generic 'jes' for normal uncertainty
            funcs.append(func)
            jettype = _checkConsistency(jettype, info[4])

        if campaign is None:
            raise Exception('Unable to determine production campaign of JECs!')
        else:
            self._campaign = campaign

        if dataera is None:
            raise Exception('Unable to determine data era of JECs!')
        else:
            self._dataera = dataera

        if datatype is None:
            raise Exception('Unable to determine if JECs are for MC or Data!')
        else:
            self._datatype = datatype

        if len(levels) == 0:
            raise Exception('No levels provided?')
        else:
            self._levels = levels
            self._funcs = funcs

        if jettype is None:
            raise Exception('Unable to determine type of jet to correct!')
        else:
            self._jettype = jettype
        # now we setup the call signature for this factorized JEC
        self._signature = []
        for func in self._funcs:
            sig = func.signature
            for input in sig:
                if input not in self._signature:
                    self._signature.append(input)

    @property
    def signature(self):
        """ list the necessary jet properties that must be input to this function """
        return self._signature

    @property
    def levels(self):
        """ list the different sources of uncertainty """
        return self._levels

    def __repr__(self):
        out = 'campaign   : %s\n' % (self._campaign)
        out += 'data era   : %s\n' % (self._dataera)
        out += 'data type  : %s\n' % (self._datatype)
        out += 'jet type   : %s\n' % (self._jettype)
        out += 'levels     : %s\n' % (','.join(self._levels))
        out += 'signature  : (%s)\n' % (','.join(self._signature))
        return out

    def getUncertainty(self, **kwargs):
        """
            Returns the set of uncertainties for all input jets for all the levels (== sources)
            use like:
            juncs = uncertainty.getUncertainty(JetProperty1=jet.property1,...)
            'juncs' will be formatted like [('SourceName', [[up_val down_val]_jet1 ... ]), ...]
            in a zip iterator
        """
        uncs = []
        for i, func in enumerate(self._funcs):
            sig = func.signature
            args = []
            for input in sig:
                args.append(kwargs[input])
            uncs.append(func(*tuple(args)))
        return zip(self._levels, uncs)



from fnal_column_analysis_tools.lookup_tools.lookup_base import lookup_base

from fnal_column_analysis_tools.util import awkward
from fnal_column_analysis_tools.util import numpy as np
from copy import deepcopy


class dense_lookup(lookup_base):
    def __init__(self, values, dims, feval_dim=None):
        super(dense_lookup, self).__init__()
        self._dimension = 0
        whattype = type(dims)
        if whattype == np.ndarray:
            self._dimension = 1
        else:
            self._dimension = len(dims)
        if self._dimension == 0:
            raise Exception('Could not define dimension for {}'.format(whattype))
        self._axes = deepcopy(dims)
        self._feval_dim = None
        vals_are_strings = ('string' in values.dtype.name or
                            'str' in values.dtype.name or
                            'unicode' in values.dtype.name or
                            'bytes' in values.dtype.name)  # ....
        if not isinstance(values, np.ndarray):
            raise TypeError("values is not a numpy array, but %r" % type(values))
        if vals_are_strings:
            raise Exception("dense_lookup cannot handle string values!")
        self._values = deepcopy(values)

    def _evaluate(self, *args):
        indices = []
        for arg in args:
            if type(arg) == awkward.JaggedArray:
                raise Exception('JaggedArray in inputs')
        if self._dimension == 1:
            indices.append(np.clip(np.searchsorted(self._axes, args[0], side='right') - 1, 0, self._values.shape[0] - 1))
        else:
            for dim in range(self._dimension):
                indices.append(np.clip(np.searchsorted(self._axes[dim], args[dim], side='right') - 1, 0, self._values.shape[dim] - 1))
        return self._values[tuple(indices)]

    def __repr__(self):
        myrepr = "{} dimensional histogram with axes:\n".format(self._dimension)
        temp = ""
        if self._dimension == 1:
            temp = "\t1: {}\n".format(self._axes)
        else:
            temp = "\t1: {}\n".format(self._axes[0])
        for idim in range(1, self._dimension):
            temp += "\t{}: {}\n".format(idim + 1, self._axes[idim])
        myrepr += temp
        return myrepr



from fnal_column_analysis_tools import hist, processor
from copy import deepcopy
from concurrent.futures import as_completed
from collections.abc import Sequence

from tqdm import tqdm
import cloudpickle as cpkl
import lz4.frame as lz4f
import numpy as np
import pandas as pd
from collections import defaultdict

from parsl.app.app import python_app


@python_app
def coffea_pyapp(dataset, fn, treename, chunksize, index, procstr):
    import uproot
    import cloudpickle as cpkl
    import lz4.frame as lz4f
    from fnal_column_analysis_tools import hist, processor
    from fnal_column_analysis_tools.processor.accumulator import accumulator

    # instrument xrootd source
    if not hasattr(uproot.source.xrootd.XRootDSource, '_read_real'):

        def _read(self, chunkindex):
            self.bytesread = getattr(self, 'bytesread', 0) + self._chunkbytes
            return self._read_real(chunkindex)

        uproot.source.xrootd.XRootDSource._read_real = uproot.source.xrootd.XRootDSource._read
        uproot.source.xrootd.XRootDSource._read = _read

    processor_instance = cpkl.loads(lz4f.decompress(procstr))

    afile = uproot.open(fn)
    tree = None
    if isinstance(treename, str):
        tree = afile[treename]
    elif isinstance(treename, Sequence):
        for name in reversed(treename):
            if name in afile:
                tree = afile[name]
    else:
        raise Exception('treename must be a str or Sequence but is a %s!' % repr(type(treename)))

    if tree is None:
        raise Exception('No tree found, out of possible tree names: %s' % repr(treename))

    df = processor.LazyDataFrame(tree, chunksize, index, flatten=True)
    df['dataset'] = dataset

    vals = processor_instance.process(df)
    vals['_bytesread'] = accumulator(afile.source.bytesread if isinstance(afile.source, uproot.source.xrootd.XRootDSource) else 0)

    return vals, tree.numentries, dataset


class ParslExecutor(object):

    def __init__(self):
        self._counts = {}

    @property
    def counts(self):
        return self._counts

    def __call__(self, dfk, items, processor_instance, output, unit='items', desc='Processing'):
        procstr = lz4f.compress(cpkl.dumps(processor_instance))

        nitems = len(items)
        ftr_to_item = set()
        for dataset, fn, treename, chunksize, index in items:
            if dataset not in self._counts:
                self._counts[dataset] = 0
            ftr_to_item.add(coffea_pyapp(dataset, fn, treename, chunksize, index, procstr))

        for ftr in tqdm(as_completed(ftr_to_item), total=nitems, unit='items', desc='Processing'):
            ftrhist, nentries, dataset = ftr.result()
            self._counts[dataset] += nentries
            output.add(ftrhist)


parsl_executor = ParslExecutor()



from fnal_column_analysis_tools.analysis_objects.JaggedCandidateArray import JaggedCandidateArray
from fnal_column_analysis_tools.analysis_objects.JaggedCandidateArray import JaggedTLorentzVectorArray



from .FactorizedJetCorrector import FactorizedJetCorrector
from .JetResolution import JetResolution
from .JetResolutionScaleFactor import JetResolutionScaleFactor
from .JetCorrectionUncertainty import JetCorrectionUncertainty
from .JetTransformer import JetTransformer
# from .MetUncertaintyCalculator import calculateType1MetXY



import parsl
import os
import shutil
import os.path as osp
from parsl.app.app import python_app, bash_app
from parsl.configs.local_threads import config

from parsl.providers import SlurmProvider
from parsl.channels import LocalChannel
from parsl.launchers import SrunLauncher
from parsl.config import Config
from parsl.executors import HighThroughputExecutor

from parsl.addresses import address_by_hostname

x509_proxy = 'x509up_u%s' % (os.getuid())


def slurm_config(cores_per_job=16, mem_per_core=2048,
                 jobs_per_worker=1,
                 initial_workers=4, max_workers=8,
                 work_dir=os.environ['WORK'],
                 grid_proxy_dir='/tmp',
                 partition='',
                 walltime='02:00:00',
                 htex_label='coffea_parsl_slurm_htex'):

    shutil.copy2(osp.join(grid_proxy_dir, x509_proxy), osp.join(work_dir, x509_proxy))

    wrk_init = '''
    export XRD_RUNFORKHANDLER=1
    export X509_USER_PROXY=%s
    ''' % (os.join(work_dir, x509_proxy))

    sched_opts = '''
    #SBATCH --cpus-per-task=%d
    #SBATCH --mem-per-cpu=%d
    ''' % (cores_per_job, mem_per_core, )

    slurm_htex = Config(
        executors=[
            HighThroughputExecutor(
                label=htex_label,
                address=address_by_hostname(),
                prefetch_capacity=0,
                max_workers=cores_per_job,
                provider=SlurmProvider(
                    channel=LocalChannel(),
                    launcher=SrunLauncher(),
                    init_blocks=initial_workers,
                    max_blocks=max_workers,
                    nodes_per_block=jobs_per_worker,
                    partition=partition,
                    scheduler_options=sched_opts,   # Enter scheduler_options if needed
                    worker_init=wrk_init,         # Enter worker_init if needed
                    walltime=walltime
                ),
            )
        ],
        strategy=None,
    )

    return slurm_htex



from __future__ import division
import time
import uproot
import multiprocessing
import copy

import threading

xrdconfig = {'timeout': None, 'chunkbytes': 1 << 14, 'limitbytes': 1 << 24}


class JaggedStripeDummy(object):
    """
        Such a shame to de-jaggedize and then later re-jaggedize, oh well
    """
    def __init__(self, children):
        self._children = {k: v.content for k, v in children.items()}
        self.count = next(iter(children.values())).counts

    def __repr__(self):
        return "<Jagged object (children=%s) at %08x>" % (",".join(self._children.keys()), id(self))

    def __getattr__(self, key):
        if key in self._children:
            return self._children[key]


class EventsDummy(object):
    def __init__(self, arrays, nevents):
        self._arrays = arrays
        self.nevents = nevents

    def __getattr__(self, key):
        if key in self._arrays:
            return self._arrays[key]


class JobDummy(object):
    def __init__(self, fname, treename, user_params, eventsdef, worker):
        self._fname = fname
        self._treename = treename
        self._user_params = copy.deepcopy(user_params)
        if 'hists' in self._user_params:
            for h in self._user_params['hists'].values():
                h.clear()
        self._eventsdef = eventsdef
        self._worker = worker
        self._out = None

    def __getitem__(self, key):
        return self._user_params[key]

    def send(self, **kwargs):
        if self._out is not None:
            raise Exception("UprootJob only supports one job.send call per worker run call")
        self._out = kwargs

    def __call__(self):
        file = uproot.open(self._fname, xrootdsource=xrdconfig)
        tree = file[self._treename]
        nevents = tree.numentries
        # FIXME: big hack for reading datasets that don't have all columns
        #   up to the worker to handle missing columns, and I'm not sure how striped
        #   will handle this (it doesn't currently)
        bnames = [b for b in self._eventsdef.arraynames() if b in tree]
        arrays = tree.arrays(branches=bnames, namedecode='ascii')
        eventsdummy = self._eventsdef.make_dummy(arrays, nevents)
        self._worker.run(eventsdummy, self)
        return (eventsdummy.nevents, self._out)


class UprootEvents(object):
    def __init__(self, columns, jagged_separator='_'):
        self._flatbranches = [c for c in columns if "." not in c]
        self._jaggedbranches = [c.split(".") for c in columns if "." in c]
        if any(len(b) > 2 for b in self._jaggedbranches):
            raise Exception("Not designed to work for jaggedness larger than 1")
        self._sep = jagged_separator
        self._arraynames = self._flatbranches + [self._sep.join(b) for b in self._jaggedbranches]

    def arraynames(self):
        return self._arraynames

    def make_dummy(self, arrays, nevents):
        parents = set(b[0] for b in self._jaggedbranches)
        for parent in parents:
            children = [self._sep.join(b) for b in self._jaggedbranches if b[0] == parent]
            childnames = [b[1] for b in self._jaggedbranches if b[0] == parent]
            child_dict = {}
            for name, bname in zip(childnames, children):
                child_dict[name] = arrays.pop(bname)
            arrays[parent] = JaggedStripeDummy(child_dict)
        return EventsDummy(arrays, nevents)


class UprootJob(object):
    def __init__(self, dataset, filelist, treename, worker_class, user_callback, user_params):
        self._dataset = dataset
        self._filelist = filelist
        self._treename = treename
        self._worker = worker_class()
        self._user_callback = user_callback
        self._user_params = user_params

        self._eventsdef = UprootEvents(self._worker.Columns)
        self.EventsProcessed = 0
        self._filesProcessed = 0

    def work(self, fname):
        pass

    def run(self, workers=0, progress=True, threaded=True):
        if threaded:
            thread = threading.Thread(target=self._run, args=(workers, progress))
            thread.start()
            return thread
        self._run(workers, progress)

    def _run(self, workers, progress):
        self.TStart = time.time()

        pool = None
        if workers > 0:
            pool = multiprocessing.Pool(processes=workers)

        pbar = None
        if progress:
            from IPython.display import display
            import ipywidgets

            lbl = ipywidgets.Label(self._dataset, layout=ipywidgets.Layout(width="70%"))
            info = ipywidgets.Label("- kevt/s", layout=ipywidgets.Layout(width="10%", display='flex', justify_content='flex-end'))
            pbar = ipywidgets.IntProgress(min=0, max=len(self._filelist), layout=ipywidgets.Layout(width="17%"))
            cancel = ipywidgets.Button(tooltip='Abort processing', icon='times', button_style='danger', layout=ipywidgets.Layout(width="3%"))

            def abort(b):
                if pool is not None:
                    pool.terminate()
                b.disabled = True
            cancel.on_click(abort)
            bar = ipywidgets.HBox([lbl, info, pbar, cancel])
            display(bar)

        try:
            jobs = (JobDummy(fname, self._treename, self._user_params, self._eventsdef, self._worker) for fname in self._filelist)
            if pool is not None:
                res = set(pool.apply_async(job) for job in jobs)
                while True:
                    finished = set()
                    for r in res:
                        if r.ready():
                            if not r.successful():
                                raise r.get()
                            out = r.get()
                            with self._user_callback.lock:
                                self._user_callback.on_streams_update(*out)
                            self.EventsProcessed += out[0]
                            self._filesProcessed += 1
                            finished.add(r)
                    res -= finished
                    finished = None
                    if progress:
                        pbar.value = self._filesProcessed
                        info.value = "{:>5.0f} kevt/s".format(self.EventsProcessed / (time.time() - self.TStart) / 1000)
                        if cancel.disabled:
                            break
                    if len(res) == 0:
                        if progress:
                            cancel.disabled = True
                        break
                    time.sleep(0.2)
            else:
                for job in jobs:
                    out = job()
                    with self._user_callback.lock:
                        self._user_callback.on_streams_update(*out)
                    self.EventsProcessed += out[0]
                    self._filesProcessed += 1
                    if progress:
                        pbar.value = self._filesProcessed
                        info.value = "{:>5.0f} kevt/s".format(self.EventsProcessed / (time.time() - self.TStart) / 1000)
                        if cancel.disabled:
                            break
                if progress:
                    cancel.disabled = True
        except Exception:
            if pool is not None:
                pool.terminate()
            if progress:
                cancel.disabled = True
                info.value = "Exception"
            raise

        pool = None
        self.TFinish = time.time()



from ..util import awkward
import uproot_methods
from uproot_methods.classes.TLorentzVector import JaggedArrayMethods, ArrayMethods, TLorentzVectorArray, TLorentzVector


class Dangerousness(object):
    # NB: This is dangerous
    def __getattr__(self, what):
        if what in self.columns:
            return self[what]
        thewhat = getattr(super(Dangerousness, self), what)
        return self.fromjagged(thewhat)


class SaiyanCommonMethods(Dangerousness, ArrayMethods):
    def match(self, cands, value):
        if isinstance(self, awkward.JaggedArray):
            combinations = self.cross(cands, True)
            mask = (combinations.i0.delta_r(combinations.i1) < value)
        else:
            mask = self.delta_r(cands) < value
        return mask.any()

    def closest(self, cands):
        if isinstance(self, awkward.JaggedArray):
            combinations = self.cross(cands, nested=True)
            if ((~(combinations.i0.eta == 0).flatten().flatten().all()) | (~(combinations.i1.eta == 0).flatten().flatten().all())):
                criterium = combinations.i0.delta_phi(combinations.i1)
            else:
                criterium = combinations.i0.delta_r(combinations.i1)
        else:
            if ((~(self.eta == 0).flatten().flatten().all()) | (~(cands.eta == 0).flatten().flatten().all())):
                criterium = self.delta_phi(cands)
            else:
                criterium = self.delta_r(cands)
        return cands[criterium.argmin()]


class JaggedSaiyanArrayMethods(SaiyanCommonMethods, JaggedArrayMethods):
    def pairs(self, nested=False):
        outs = super(JaggedSaiyanArrayMethods, self).pairs(nested)
        return self.fromjagged(outs)

    def distincts(self, nested=False):
        outs = super(JaggedSaiyanArrayMethods, self).distincts(nested)
        return self.fromjagged(outs)

    def cross(self, other, nested=False):
        outs = super(JaggedSaiyanArrayMethods, self).cross(other, nested)
        return self.fromjagged(outs)


class SaiyanArray(SaiyanCommonMethods, TLorentzVectorArray):
    def _initObjectArray(self, table):
        self.awkward.ObjectArray.__init__(self, table, lambda row: TLorentzVector(row["fX"], row["fY"], row["fZ"], row["fE"]))


def Initialize(items):
    argkeys = items.keys()
    p4 = None

    if 'p4' in argkeys:
        p4 = items['p4']
        if not isinstance(p4, uproot_methods.TLorentzVectorArray):
            p4 = SaiyanArray.from_cartesian(p4[:, 0], p4[:, 1],
                                            p4[:, 2], p4[:, 3])

    elif 'pt' in argkeys and 'eta' in argkeys and 'phi' in argkeys and 'mass' in argkeys:
        temp = SaiyanArray.from_ptetaphim(items['pt'], items['eta'],
                                          items['phi'], items['mass'])._to_cartesian()
        p4 = SaiyanArray.from_cartesian(temp.x, temp.y, temp.z, temp.t)
        del items['pt']
        del items['eta']
        del items['phi']
        del items['mass']

    elif 'pt' in argkeys and 'eta' in argkeys and 'phi' in argkeys and 'energy' in argkeys:
        p4 = SaiyanArray.from_ptetaphi(items['pt'], items['eta'],
                                       items['phi'], items['energy'])
        del items['pt']
        del items['eta']
        del items['phi']
        del items['energy']

    elif 'px' in argkeys and 'py' in argkeys and 'pz' in argkeys and 'mass' in argkeys:
        p4 = SaiyanArray.from_xyzm(items['px'], items['py'],
                                   items['pz'], items['mass'])
        del items['px']
        del items['py']
        del items['pz']
        del items['mass']

    elif 'pt' in argkeys and 'phi' in argkeys and 'pz' in argkeys and 'energy' in argkeys:
        p4 = SaiyanArray.from_cylindrical(items['pt'], items['phi'],
                                          items['pz'], items['energy'])
        del items['pt']
        del items['phi']
        del items['pz']
        del items['energy']

    elif 'px' in argkeys and 'py' in argkeys and 'pz' in argkeys and 'energy' in argkeys:
        p4 = SaiyanArray.from_cartesian(items['px'], items['py'],
                                        items['pz'], items['energy'])
        del items['px']
        del items['py']
        del items['pz']
        del items['energy']

    elif 'p' in argkeys and 'theta' in argkeys and 'phi' in argkeys and 'energy' in argkeys:
        p4 = SaiyanArray.from_spherical(items['p'], items['theta'],
                                        items['phi'], items['energy'])
        del items['p']
        del items['theta']
        del items['phi']
        del items['energy']

    elif 'p3' in argkeys and 'energy' in argkeys:
        p4 = SaiyanArray.from_p3(items['p3'], items['energy'])
        del items['p3']
        del items['energy']

    try:
        p4
    except NameError:
        out = awkward.Table()
    else:
        out = p4

    for name, value in items.items():
        out[name] = value

    if isinstance(out, awkward.JaggedArray):
        out = JaggedSaiyanArrayMethods.fromjagged(out)
    else:  # p4 is not None and not isinstance(out, awkward.JaggedArray):
        try:
            p4
        except NameError:
            out.__class__ = type("Event", (out.__class__,
                                           Dangerousness), {})
        else:
            out.__class__ = type("Object", (out.__class__,
                                            SaiyanCommonMethods), {})

    return out



from fnal_column_analysis_tools import hist, processor
from copy import deepcopy
from concurrent.futures import ThreadPoolExecutor, as_completed

from tqdm import tqdm
import _pickle as pkl
import lz4.frame as lz4f
import numpy as np
import pandas as pd

import pyspark
import pyspark.sql.functions as fn
from pyspark.sql.types import BinaryType, StringType, StructType, StructField

from jinja2 import Environment, PackageLoader, select_autoescape

lz4_clevel = 1

# this is a UDF that takes care of summing histograms across
# various spark results where the outputs are histogram blobs
@fn.pandas_udf(BinaryType(), fn.PandasUDFType.GROUPED_AGG)
def agg_histos(df):
    global processor_instance, lz4_clevel
    goodlines = df[df.str.len() > 0]
    if goodlines.size == 1:  # short-circuit trivial aggregations
        return goodlines[0]
    outhist = processor_instance.accumulator.identity()
    for line in goodlines:
        outhist.add(pkl.loads(lz4f.decompress(line)))
    return lz4f.compress(pkl.dumps(outhist), compression_level=lz4_clevel)


@fn.pandas_udf(StructType([StructField('histos', BinaryType(), True)]),
               fn.PandasUDFType.GROUPED_MAP)
def reduce_histos(df):
    histos = df['histos']
    mask = (histos.str.len() > 0)
    outhist = processor_instance.accumulator.identity()
    for line in histos[mask]:
        outhist.add(pkl.loads(lz4f.decompress(line)))
    return pd.DataFrame(data={'histos': np.array([lz4f.compress(pkl.dumps(outhist))], dtype='O')})


class SparkExecutor(object):
    _template_name = 'spark.py.tmpl'

    def __init__(self):
        self._cacheddfs = None
        self._rawresults = None
        self._counts = None
        self._env = Environment(loader=PackageLoader('fnal_column_analysis_tools.processor',
                                                     'templates'),
                                autoescape=select_autoescape(['py'])
                                )

    @property
    def counts(self):
        return self._counts

    def __call__(self, spark, dfslist, theprocessor, output, thread_workers,
                 unit='datasets', desc='Processing'):
        # processor needs to be a global
        global processor_instance, coffea_udf
        processor_instance = theprocessor
        # get columns from processor
        columns = processor_instance.columns
        cols_w_ds = ['dataset'] + columns
        # make our udf
        tmpl = self._env.get_template(self._template_name)
        render = tmpl.render(cols=columns)
        exec(render)

        # cache the input datasets if it's not already done
        if self._cacheddfs is None:
            self._cacheddfs = {}
            self._counts = {}
            # go through each dataset and thin down to the columns we want
            for ds, (df, counts) in dfslist.items():
                self._cacheddfs[ds] = df.select(*cols_w_ds).cache()
                self._counts[ds] = counts

        with ThreadPoolExecutor(max_workers=thread_workers) as executor:
            future_to_ds = {}
            for ds, df in self._cacheddfs.items():
                future_to_ds[executor.submit(self._launch_analysis, df, coffea_udf, cols_w_ds)] = ds
            # wait for the spark jobs to come in
            self._rawresults = {}
            for future in tqdm(as_completed(future_to_ds),
                               total=len(future_to_ds),
                               desc=desc, unit=unit):
                self._rawresults[future_to_ds[future]] = future.result()

        for ds, bitstream in self._rawresults.items():
            if bitstream is None:
                raise Exception('No pandas dataframe returns from spark in dataset: %s, something went wrong!' % ds)
            if bitstream.empty:
                raise Exception('The histogram list returned from spark is empty in dataset: %s, something went wrong!' % ds)
            bits = bitstream[bitstream.columns[0]][0]
            output.add(pkl.loads(lz4f.decompress(bits)))

    def _launch_analysis(self, df, udf, columns):
        histo_map_parts = (df.rdd.getNumPartitions() // 20) + 1
        return df.select(udf(*columns).alias('histos')) \
                 .withColumn('hpid', fn.spark_partition_id() % histo_map_parts) \
                 .repartition(histo_map_parts, 'hpid') \
                 .groupBy('hpid').apply(reduce_histos) \
                 .groupBy().agg(agg_histos('histos')) \
                 .toPandas()


spark_executor = SparkExecutor()



from six import with_metaclass
from abc import ABCMeta, abstractmethod


class ProcessorABC(with_metaclass(ABCMeta)):
    '''
    ABC for a generalized processor
    The various data delivery mechanisms (spark, striped, uproot, uproot+futures, condor, ...)
    could receive such an object and the appropriate metadata to deliver dataframes to it.
    It is expected that the entire processor object can be cloudpickle-able

    No attempt should be made to track state inside an instance of ProcessorABC, it is to be
    treated simply as a bundle of methods.  The only exception is the read-only accumulator
    property.
    '''
    @property
    @abstractmethod
    def accumulator(self):
        '''
        Read-only accumulator object, as defined in e.g. the initializer of the concrete class.
        '''
        pass

    @abstractmethod
    def process(self, df):
        '''
        Processes a single DataFrame, and returns a filled accumulator object
        which can be initialized via self.accumulator.identity()
        '''
        pass

    @abstractmethod
    def postprocess(self, accumulator):
        '''
        Do any final processing on the resulting accumulator object, and return it
        '''
        pass



from fnal_column_analysis_tools.util import awkward
from fnal_column_analysis_tools.util import numpy as np


class lookup_base(object):
    # base class for all objects that do some sort of value or function lookup
    def __init__(self):
        pass

    def __call__(self, *args):
        inputs = list(args)
        offsets = None
        # TODO: check can use offsets (this should always be true for striped)
        # Alternatively we can just use starts and stops
        for i in range(len(inputs)):
            if isinstance(inputs[i], awkward.JaggedArray):
                if offsets is not None and offsets.base is not inputs[i].offsets.base:
                    if type(offsets) is int:
                        raise Exception('Do not mix JaggedArrays and numpy arrays when calling derived class of lookup_base')
                    elif type(offsets) is np.ndarray and offsets.base is not inputs[i].offsets.base:
                        raise Exception('All input jagged arrays must have a common structure (offsets)!')
                offsets = inputs[i].offsets
                inputs[i] = inputs[i].content
            elif isinstance(inputs[i], np.ndarray):
                if offsets is not None:
                    if type(offsets) is np.ndarray:
                        raise Exception('do not mix JaggedArrays and numpy arrays when calling a derived class of lookup_base')
                offsets = -1
        retval = self._evaluate(*tuple(inputs))
        if offsets is not None and type(offsets) is not int:
            retval = awkward.JaggedArray.fromoffsets(offsets, retval)
        return retval

    def _evaluate(self, *args):
        raise NotImplementedError



# if (jetPtJESUp > 20) {
#    MetXCorrjesUp += -1 * (thisJetJESUp.Px() - thisJet.Px());
#    MetYCorrjesUp += -1 * (thisJetJESUp.Py() - thisJet.Py());
# }
# if (jetPtJESDown > 20) {
#    MetXCorrjesDown += -1 * (thisJetJESDown.Px() - thisJet.Px());
#    MetYCorrjesDown += -1 * (thisJetJESDown.Py() - thisJet.Py());
# }
# if (jetPtJERUp > 20) {
#    MetXCorrjerUp += -1 * (thisJetJERUp.Px() - thisJet.Px());
#    MetYCorrjerUp += -1 * (thisJetJERUp.Py() - thisJet.Py());
# }
# if (jetPtJERDown > 20) {
#    MetXCorrjerDown += -1 * (thisJetJERDown.Px() - thisJet.Px());
#    MetYCorrjerDown += -1 * (thisJetJERDown.Py() - thisJet.Py());
# }

from ..analysis_objects.JaggedCandidateArray import JaggedCandidateArray

from fnal_column_analysis_tools.util import numpy as np
from fnal_column_analysis_tools.util import awkward


# returns corrected MET.x, corrected MET.y
def calculateType1MetXY(pfmet_in, correctedAK4Jets):
    pass
